/*
 * tmos - i51-slave.c
 *
 * ModBUS RTU slave for i51
 *
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/port.h>
#include <tmos/crc16.h>
#include <asm/8052.h>
#include <asm/system.h>
#include <asm/task.h>
#include <tmos/errno.h>
#include <tmos/net.h>
#include <asm/irqs.h>
#include <limits.h>
#include <asm/aduc841.h>

#ifndef dir_tx
#define dir_tx()	do { } while (0)
#endif
#ifndef dir_rx
#define dir_rx()	do { } while (0)
#endif

#define freq_to_timer(x) \
	(65536 - CONFIG_OSCILLATOR / (CONFIG_TIMER_DIVISOR * x))

/* 3.5 character inter-frame gap */
#if CONFIG_BAUDRATE_9600
#define T35 		freq_to_timer(9600 / 35)
#define TSTOP 		freq_to_timer(9600 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_9600
#elif CONFIG_BAUDRATE_19200
#define T35 		freq_to_timer(19200 / 35)
#define TSTOP 		freq_to_timer(19200 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_19200
#elif CONFIG_BAUDRATE_38400
#define T35 		freq_to_timer(38400 / 35)
#define TSTOP 		freq_to_timer(38400 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_38400
#elif CONFIG_BAUDRATE_57600
#define T35 		freq_to_timer(57600 / 35)
#define TSTOP 		freq_to_timer(57600 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_57600
#elif CONFIG_BAUDRATE_115200
#define T35 		freq_to_timer(115200 / 35)
#define TSTOP 		freq_to_timer(115200 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_115200
#elif CONFIG_BAUDRATE_125000
#define T35 		freq_to_timer(125000 / 35)
#define TSTOP 		freq_to_timer(125000 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_125000
#elif CONFIG_BAUDRATE_230400
#define T35 		freq_to_timer(230400 / 35)
#define TSTOP 		freq_to_timer(230400 / 15)
#define DEFAULT_BAUDRATE	NET_SPEED_230400
#else
#error "rtu: unsupported serial baudrate"
#endif

#ifdef CONFIG_RTU_SLAVE_ADDR_RUNTIME
static uint8_t rtu_addr = CONFIG_RTU_SLAVE_ADDR;
#define rtu_address()	rtu_addr
#else
#define rtu_address()	CONFIG_RTU_SLAVE_ADDR
#endif

#define rtu_timer_start(x) \
	do { \
		TR1 = 0; \
		TH1 = (x) >> 8; \
		TL1 = (x) & 0xff; \
		TR1 = 1; \
	} while (0)

#define __rtu_timer_disable() \
	do { \
		TR1 = 0; \
	} while (0)

#define __rtu_timer_enable() \
	do { \
		TR1 = 1; \
	} while (0)

#define __rtu_timer_start(x) \
	do { \
		TH1 = (x) >> 8; \
		TL1 = (x) & 0xff; \
	} while (0)

#ifdef CONFIG_RTU_SLAVE_BAUDRATE_RUNTIME

static unsigned char baudrate;

#define rtu_timer_start_interframe() \
	do { \
		__rtu_timer_disable(); \
		switch (baudrate) { \
		case NET_SPEED_9600: \
			__rtu_timer_start(freq_to_timer(9600 / 35)); \
			break; \
		case NET_SPEED_19200: \
			__rtu_timer_start(freq_to_timer(19200 / 35)); \
			break; \
		case NET_SPEED_38400: \
			__rtu_timer_start(freq_to_timer(38400 / 35)); \
			break; \
		case NET_SPEED_57600: \
			__rtu_timer_start(freq_to_timer(57600 / 35)); \
			break; \
		case NET_SPEED_115200: \
			__rtu_timer_start(freq_to_timer(115200 / 35)); \
			break; \
		case NET_SPEED_125000: \
			__rtu_timer_start(freq_to_timer(125000 / 35)); \
			break; \
		case NET_SPEED_230400: \
			__rtu_timer_start(freq_to_timer(230400 / 35)); \
			break; \
		} \
		__rtu_timer_enable(); \
	} while (0)

#define rtu_timer_start_stop() \
	do { \
		__rtu_timer_disable(); \
		switch (baudrate) { \
		case NET_SPEED_9600: \
			__rtu_timer_start(freq_to_timer(9600 / 15)); \
			break; \
		case NET_SPEED_19200: \
			__rtu_timer_start(freq_to_timer(19200 / 15)); \
			break; \
		case NET_SPEED_38400: \
			__rtu_timer_start(freq_to_timer(38400 / 15)); \
			break; \
		case NET_SPEED_57600: \
			__rtu_timer_start(freq_to_timer(57600 / 15)); \
			break; \
		case NET_SPEED_115200: \
			__rtu_timer_start(freq_to_timer(115200 / 15)); \
			break; \
		case NET_SPEED_125000: \
			__rtu_timer_start(freq_to_timer(125000 / 15)); \
			break; \
		case NET_SPEED_230400: \
			__rtu_timer_start(freq_to_timer(230400 / 15)); \
			break; \
		} \
		__rtu_timer_enable(); \
	} while (0)

#else
#define rtu_timer_start_interframe() rtu_timer_start(T35)
#define rtu_timer_start_stop() rtu_timer_start(TSTOP)
#endif

#define rtu_timer_interrupt_disable() do { ET1 = 0; } while (0)
#define rtu_timer_interrupt_enable() do { ET1 = 1; } while (0)
#define serial_interrupt_disable() do { ES = 0; } while (0)
#define serial_interrupt_enable() do { ES = 1; } while (0)

#define RTU_MTU		CONFIG_RTU_MTU

#define TXP_END 0xff

#ifdef CONFIG_KERNEL_DATA
static uint8_t __idata msg_buffer[RTU_MTU];
#else
static uint8_t __kernel msg_buffer[RTU_MTU];
#endif
static uint8_t __kernel msg_size = 0;
static uint8_t __kernel msg_txp = TXP_END;	/* transmit pointer */
static __bit rx_lock = 0;

#define recv_char(c) \
	do { \
		rtu_timer_start_interframe(); \
		if (msg_size < RTU_MTU) { \
			msg_buffer[msg_size] = c; \
			msg_size++; \
		} \
	} while (0)

void timer1_irq(void) __interrupt(3) __using(TIMER1_BANK)
{
	serial_interrupt_disable();
	if (rx_lock) {
		dir_rx();
		msg_size = 0;
		rx_lock = 0;
	} else if (msg_size >= 3) {
		if ((msg_buffer[0] == rtu_address()) || !msg_buffer[0]) {
			rx_lock = 1;
			msg_txp = TXP_END;
		} else {
			msg_size = 0;
		}
	}
	TR1 = 0;
	serial_interrupt_enable();
}

void ser_irq(void) __interrupt(SERIAL_INTERRUPT) __using(I51_SERIAL_BANK)
{
	unsigned char tmp;

	if (RI) {
		RI = 0;
		tmp = SBUF;

		if (!rx_lock)
			recv_char(tmp);
	}

	if (TI) {
		TI = 0;
		if (msg_txp < msg_size) {
			__tx_send_raw(msg_buffer[msg_txp]);
			msg_txp++;
		} else {
			rtu_timer_start_stop();
		}
	}
}

char rtu_serial_i51_ioctl(unsigned char cmd, void __kernel *arg) __reentrant
{
	char ret = -EINVAL;
	unsigned char x;

	switch (cmd) {
#ifdef CONFIG_RTU_SLAVE_ADDR_RUNTIME
	case NET_IOCTL_SET_ADDR_BYTE:
		rtu_addr = *((uint8_t __kernel *) arg);
		ret = 0;
		break;
#endif
#ifdef CONFIG_RTU_SLAVE_BAUDRATE_RUNTIME
	case NET_IOCTL_SET_SPEED:
		x = *((uint8_t __kernel *) arg);
		ret = 0;
		switch (x) {
#ifdef B9600
		case NET_SPEED_9600:
			ser_set_spd(B9600);
			baudrate = x;
			break;
#endif
#ifdef B19200
		case NET_SPEED_19200:
			ser_set_spd(B19200);
			baudrate = x;
			break;
#endif
#ifdef B38400
		case NET_SPEED_38400:
			ser_set_spd(B38400);
			baudrate = x;
			break;
#endif
#ifdef B57600
		case NET_SPEED_57600:
			ser_set_spd(B57600);
			baudrate = x;
			break;
#endif
#ifdef B115200
		case NET_SPEED_115200:
			ser_set_spd(B115200);
			baudrate = x;
			break;
#endif
#ifdef B125000
		case NET_SPEED_125000:
			ser_set_spd(B125000);
			baudrate = x;
			break;
#endif
#ifdef B230400
		case NET_SPEED_230400:
			ser_set_spd(B230400);
			baudrate = x;
			break;
#endif
		default:
			ret = -EINVAL;
		}
		break;
#endif
	}
	return ret;
}

void rtu_serial_i51_poll(void) __reentrant
{
	if (rx_lock && (msg_txp == TXP_END)) {
		net_size_t ret = 0;
		uint16_t __kernel *crc;
		struct net_msg msg;

		/* setup message */
		msg_init(&msg);
		msg_set_interface(&msg, NET_IFACE_SERIAL0);
		msg_set_buffer(&msg, (void __data *)msg_buffer);
		msg_set_head(&msg, 0);
		msg_set_tail(&msg, msg_size - 2);
		msg_set_size(&msg, RTU_MTU - 2);

		crc = (uint16_t __kernel *)(msg_buffer + msg_size - 2);
		if (le16_to_cpu(*crc) == crc16_kernel(0xffff,
					(void __kernel *) msg_buffer,
					msg_size - 2)) {
			ret = net_xmit(NET_PROTO_MODBUS, &msg);
			if (ret) {
				msg_size = ret + 2;
				crc = (uint16_t __kernel *)(msg_buffer + ret);
				*crc = cpu_to_le16(crc16_kernel(0xffff,
							(void __kernel *)
							msg_buffer, ret));
			}
		}

		serial_interrupt_disable();
		if (ret) {
			dir_tx();
			msg_txp = 0;

			/*
			 * work around for sdcc bug. sdcc just inlines
			 * assembler code without storing msg_txp.
			 */
			barrier();

			tx_send_raw(msg_buffer[msg_txp]);
			msg_txp++;
		} else {
			msg_size = 0;
			rx_lock = 0;
		}
		serial_interrupt_enable();
	}
}

static void rtu_slave_i51_init(void) __reentrant
{
	TMOD = (TMOD & ~0xf0) | 0x10;
	TF1 = 0;
	TR1 = 1;
	ET1 = 1;
	dir_rx();

#ifdef CONFIG_RTU_SLAVE_BAUDRATE_RUNTIME
	baudrate = DEFAULT_BAUDRATE;
#endif
}

device_initcall(rtu_slave_i51_init)
